/*

    this source file is a part of 
    lp_parse - a preprocessor for MIP models
    Copyright (C) 1996  Ari Jolma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    for more information please contact 
    Ari Jolma
    Helsinki University of Technology
    Laboratory of Water Resources Engineering
    Building W
    02150 Espoo, Finland
    E-mail: Ari.Jolma@hut.fi

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "util.h"
#include "hash.h"
#include "lp_index.h"

int init_index(INDEX *index, char *name) {
  index->has_value = FALSE;
  index->value = 0;
  index->has_bounds = FALSE;
  index->min = 0;
  index->max = 0;
  index->smin = NULL;
  index->smax = NULL;
  try(init_item((ITEM *)index, name));
  return TRUE;
}

INDEX *new_index(char *name) {
  INDEX *index;
  try(index = (INDEX *)MALLOC(sizeof(INDEX)));
  ctry(init_index(index, name));
  return index;
catch:
  dispose_index(index);
  return NULL;
}

void delete_index(INDEX *index) {
  delete_item((ITEM *)index);
}

void dispose_index(INDEX *index) {
  if (index) {
    delete_index(index);
    free(index);
  }
}

INDEX *add_index(INDEX *root, INDEX *i) {
  i->next = root;
  return i;
}

void remove_index(INDEX **root, int n) {
  INDEX *tmp;
  int i = 0;
  while (root && (i < n)) {
    tmp = *root;
    *root = (*root)->next;
    dispose_index(tmp);
    i++;
  }
}

void delete_indexes(INDEX **root) {
  INDEX *tmp;
  while (*root) {
    tmp = *root;
    *root = (*root)->next;
    dispose_index(tmp);
  }
}

INDEX *search_index(INDEX *root, char *name) {  
  debug_print(170,"search index %s", name);
  while (root) {
    if (!strcmp(root->name, name)) return root;
    root = root->next;
  }
  return NULL;
}

int init_index_loop(INDEX *root, int n) {
  int i = 0;
  while (root && (i<n)) {
    if (root->min > root->max) return FALSE;
    root->has_value = TRUE;
    root->value = root->min;
    root = root->next;
    i++;
  }
  return TRUE;
}

int loop_indexes(INDEX *root, int n) {
  int i = 0;
  while (root && (i<n)) {
    root->value++;
    if (root->value > root->max) {
      root->value = root->min;
      root = root->next;
      i++;
    } else 
      return TRUE;
  }
  return FALSE;
}

void print_indexes(INDEX *root) {
  while (root) {
    printf("%s has value: %i %i has bounds: %i %i %i\n", 
	   root->name, root->has_value, root->value,
	   root->has_bounds, root->min, root->max);
    root = root->next;
  }
}


/*---------------------------------------------------------------------------*/


int init_constant(CONSTANT *constant, char *name) {
  constant->next_in_list = NULL;
  constant->has_value = FALSE;
  constant->value = 0;
  try(init_item((ITEM *)constant, name));
  return TRUE;
}

CONSTANT *new_constant(char *name) {
  CONSTANT *constant;
  if (!strchr(LETTERS,*name)) die("'%s' is not a valid name for a variable",name);
  try(constant = (CONSTANT *)MALLOC(sizeof(CONSTANT)));
  ctry(init_constant(constant, name));
  return constant;
catch:
  dispose_constant(constant);
  return NULL;
}

void delete_constant(CONSTANT *constant) {
  if (constant) {
    delete_item((ITEM *)constant);
  }
}

void dispose_constant(CONSTANT *constant) {
  if (constant) {
    delete_constant(constant);
    free(constant);
  }
}

int hash_constants(CMAP *map) {  
  CONSTANT *c = map->list, *tmp;

  return TRUE;

  if (!c) try(c = new_constant("null"));
  /*try(map->hash = new_hash());*/
  while (c) {
    /*ahash(map->hash, (ITEM *)c);*/
    tmp = c;
    c = c->next_in_list;
    tmp->next = NULL;
  }
  return TRUE;
}

void add_constant(CMAP *map, CONSTANT *c) {
  debug_print(170, "add constant %s", c->name);

  if (!map->hash) map->hash = new_hash();

  if (map->hash) {
    ahash(map->hash, (ITEM *)c);
  } /*else */

  if (!map->last_in_list) {
    map->list = c;
    map->last_in_list = c;
  } else {
    map->last_in_list->next_in_list = c;
    map->last_in_list = c;
  }
}

CONSTANT *search_constant(CMAP *map, char *name) {
  CONSTANT *c;
  debug_print(170, "search constant %s", name);
  if (map->hash) {
    return (CONSTANT *)ghash(map->hash, name);
  } else {
    c = map->list;
    while (c) {
      if (strcmp(c->name, name) == 0) break;
      c = c->next_in_list;
    }
    return c;
  }    
}

int update_constant(CMAP *map, char *name, double value) {
  CONSTANT *c = search_constant(map, name);
  if (!c) {
    try(c = new_constant(name));
    add_constant(map, c);
  } 
  c->has_value = TRUE;
  c->value = value;  
  return TRUE;
}

void print_constants(CMAP *map, int values) {
  CONSTANT *c;
  ITEM **h;
  int i,k;
  if (map->hash) {
    k = nitems(map->hash);
    h = items(map->hash, k);
    for (i = 0; i < k; i++) 
      if (((CONSTANT *)h[i])->has_value) {
	if (values)
	  printf("%s=%f\n", h[i]->name, ((CONSTANT *)h[i])->value);
	else 
	  printf("%s\n", h[i]->name);
      }
    free(h);
  } else {
    c = map->list;
    while (c) {
      if (values)
	printf("%s %i %f\n", c->name, c->has_value, c->value);
      else
	printf("%s\n", c->name);
      c = c->next_in_list;
    }
  }
}




