/*

    this source file is a part of 
    lp_parse - a preprocessor for MIP models
    Copyright (C) 1996  Ari Jolma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    for more information please contact 
    Ari Jolma
    Helsinki University of Technology
    Laboratory of Water Resources Engineering
    Building W
    02150 Espoo, Finland
    E-mail: Ari.Jolma@hut.fi

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "util.h"
#include "lp_out.h"
#include "lp_index.h"
#include "lp_tools.h"
#include "lp_parse.h"

XFILE *lps_datafile = NULL;

int lps_data_redirect(char *data, XFILE *model) {
  char path[256], *s;
  if (lps_datafile && (lps_datafile->f) && (lps_datafile->f != stdin)) 
    fclose(lps_datafile->f);
  if (strcmp(data, "?") == 0) 
    lps_datafile = NULL;
  else if (strcmp(data, "-") == 0) 
    lps_datafile = xstdin();
  else {
    if (*data == '/') 
      strncpy(path, data, 255);
    else {
      if (model && model->name) 
	strncpy(path, model->name, 255); 
      else
	*path = '\0';
      s = path;
      while (*s) s++;
      while ((s != path) && (*s != '/')) s--;
      if (*s == '/') s++;
      *s = '\0';
      strncpy(s, data, 255 - strlen(path));
    }
    try(lps_datafile = open_file(path, "r"));
  }
  return TRUE;
}

char *get_data_line() {
  char *s;
  do {        
    s = xgetline(lps_datafile);
    if (s && (*s != XEOF)) {while(*s && (strchr(WHITESPACE, *s))) s++;} 
  } while ((s) && (*s != XEOF) && (strchr("#;", *s) != NULL));
  return s;
}

int ask_lps_data(CONSTANT *C) {
  char line[256];
  do {
    printf("%s ? ", C->name);
    fgets(line,256,stdin);
    if (strcmp(line, "end") == 0) return FALSE;
    if (sscanf(line, "%lf", &(C->value)) < 1) 
      printf("input error, give 'end' to quit\n");
    else return TRUE;
  } while (TRUE);
}

char *lps_input_line = NULL;

int get_lps_data(CMAP *C) {
  CONSTANT *c;
  char *str = NULL, *tok = NULL;
  for (c = C->list; c; c = c->next_in_list) {
    if (c->has_value) continue;
    if (lps_datafile != NULL) {	
      do {
	if (!lps_input_line) str = lps_input_line = get_data_line();
	if (!lps_input_line || (*lps_input_line == XEOF)) {
	  fdie(lps_datafile, "unexpected end of data input");
	} else {
	  tok = strtok(str, WHITESPACE);
	  str = NULL;
	  if (!tok) lps_input_line = NULL;
	}
      }
      while (!tok);
      if (sscanf(tok, "%lf", &(c->value)) < 1) fdie(lps_datafile, "data input error: '%s'", tok);
      if (verbose) printf("%s=%f\n", c->name, c->value);
    } else
      try(ask_lps_data(c));
    c->has_value = TRUE;    
  }
  return TRUE;
}

#define N_TAGS 9

#define CONST_TAG "const:"
#define INDEX_TAG "index:"
#define DATA_TAG "data:"
#define VAR_TAG "var:"
#define MAX_TAG "max:"
#define MIN_TAG "min:"
#define ST_TAG "s.t."
#define INPUT_TAG "input:"
#define DEBUG_TAG "debug:"

#define CONST_SECT 1
#define INDEX_SECT 2
#define DATA_SECT 1
#define VAR_SECT 8
#define MAX_SECT 4
#define MIN_SECT 5
#define ST_SECT 6
#define INPUT_SECT 7
#define DEBUG_SECT 9

char *tags[N_TAGS] = {
  CONST_TAG,
  INDEX_TAG,
  DATA_TAG,
  VAR_TAG,
  MAX_TAG,
  MIN_TAG,
  ST_TAG,
  INPUT_TAG,
  DEBUG_TAG
};

CNR tag_numbers[N_TAGS] = {
  CONST_SECT,
  INDEX_SECT,
  DATA_SECT,
  VAR_SECT,
  MAX_SECT,
  MIN_SECT,
  ST_SECT,
  INPUT_SECT,
  DEBUG_SECT
};

CMAP data = {NULL,NULL,NULL}, var = {NULL,NULL,NULL};
INDEX *indexes = NULL;

char *get_lps_line(XFILE *in, XFILE *out, CNR *section, int *section_change) {
  char *s, *tmp, tmpc;
  CNR cnr;
  int n = N_TAGS;
  do {
    s = xgetline(in);
    if (!s || (*s == XEOF)) {
      streamer_flush(out);
      return NULL;
    }
    while (*s && strchr(WHITESPACE, *s)) s++;
  } while (strchr(";#", *s) != NULL); /* skip comment lines */
  tmp = s;
  while (*tmp && !(strchr(WHITESPACE, *tmp))) tmp++; 
  tmpc = *tmp;
  *tmp = '\0';
  cnr = getcnr(s, tags, tag_numbers, &n);
  *tmp = tmpc;
  *section_change = FALSE;
  if (cnr) {
    *section = cnr;
    s = tmp;
    *section_change = TRUE;
  }
  while (*s && strchr(WHITESPACE, *s)) s++;
  return s;
}

char *parse_lps_declarations(XFILE *in, XFILE *out, char *s, CNR *section, get_lps_data_fct gdf) {
  INDEX *i;
  int sc;
  while ((s = get_lps_line(in, out, section, &sc))) {
    debug_print(2, "%lX: parsing %s", *section, s);
    if (*section == DEBUG_SECT) {
      if (strcmp(s,"data") == 0) {
	printf("Indexes:\n");
	print_indexes(indexes);
	printf("Constants:\n");
	print_constants(&data, TRUE);
	printf("Variables:\n");
	print_constants(&var, TRUE);
	exit(1);
      } else scani(s, &verbose, NULL);
    }
    if (sc && (*section != INPUT_SECT)) {
      if (!gdf(&data)) return NULL;
    }    
    if (*s) switch (*section) {
    case INDEX_SECT:      
      try(s = parse_index(s, &i, &data, indexes, in));
      indexes = add_index(indexes, i);
      break;
    case CONST_SECT:
      try(s = parse_constant(s, &data, &data, indexes, in));
      break;
    case VAR_SECT:
      try(s = parse_constant(s, &var, &data, indexes, in));
      break;
    case INPUT_SECT:
      try(lps_data_redirect(strtok(s, " \t\r\n"), in));
      try(gdf(&data));
      break;
    case MAX_SECT:;
    case MIN_SECT:;
      return s;
    }    
  }
  return s;
}

char *parse_lps_model(XFILE *in, XFILE *out, char *s, CNR *section) {  
  int dgo, sc;
  try(hash_constants(&data));
  try(hash_constants(&var));
  do {
    debug_print(2, "%lX: parsing %s", *section, s);
    dgo = FALSE;
    if (*s) switch (*section) {    
    case DEBUG_SECT:
      scani(s, &verbose, NULL);
      break;
    case MAX_SECT:
      fprintf(out->f, "%s ", MAX_TAG);
      break;
    case MIN_SECT:
      fprintf(out->f, "%s ", MIN_TAG);
      break;
    case ST_SECT:
      break;
    }
    s = parse_block(s, &data, &var, indexes, NULL, &dgo, in, out);
    if (!s) return NULL;
    if (dgo) stream_out_char(out, ';');
  } while ((s = get_lps_line(in, out, section, &sc)));
  return s;
}

int parse_lps(XFILE *in, XFILE *out) {
  CNR section;
  char *s;
  try(s = parse_lps_declarations(in, out, s, &section, &get_lps_data));
  try(s = parse_lps_model(in, out, s, &section));
  return TRUE;
}







