/*

    this source file is a part of 
    lp_parse - a preprocessor for MIP models
    Copyright (C) 1996  Ari Jolma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    for more information please contact 
    Ari Jolma
    Helsinki University of Technology
    Laboratory of Water Resources Engineering
    Building W
    02150 Espoo, Finland
    E-mail: Ari.Jolma@hut.fi

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <math.h>
#include "util.h"
#include "lp_index.h"
#include "lp_out.h"
#include "lp_tools.h"

#define N_FCT_TAGS 4

#define IF_FCT_TAG "If"
#define SUM_FCT_TAG "Sum"
#define FORALL_FCT_TAG "ForAll"
#define INT_TAG "int"

#define IF_FCT 1
#define SUM_FCT 2
#define FORALL_FCT 3
#define INT_OP 4

char *fct_tags[N_FCT_TAGS] = {
  IF_FCT_TAG,
  SUM_FCT_TAG,
  FORALL_FCT_TAG,
  INT_TAG
};

CNR fct_tag_numbers[N_FCT_TAGS] = {
  IF_FCT,
  SUM_FCT,
  FORALL_FCT,
  INT_OP
};


int tt(char c) {
  if (c == '\0') return 0;
  if (strchr(LETTERS, c)) return LETTER;
  if (strchr(NUMBERS, c)) return NUMBER;
  return c;
}

/* scans the closing paren or bracket, returns NULL if finds '\0' before it */
char *scan_closing(char *s, XFILE *f) {
  char *expr = s, ending;
  switch (*s) {
  case PARENLEFT: ending = PARENRIGHT; break;
  case BRACKETLEFT: ending = BRACKETRIGHT; break;
  case '{': ending = '}'; break;
  default:fdie(f,"internal error in scan_closing");
  }
  s++;  
  do {
    if (*s == ending) return s;
    while (*s && !strchr(PARENS, *s)) s++;
    switch (*s) {
    case 0:
      s = NULL;
      break;
    case PARENLEFT:;
    case BRACKETLEFT:;
    case '{':
      s = scan_closing(s, f);
      if (s) {
	s++;
	break;
      }
    }    
  } while (s);
  fdie(f, "missing closing parenthesis, bracket, or brace in '%s'", expr);
}

char *skip_string(char *s, XFILE *f) {
  char *expr = s, ending;
  switch (*s) {
  case '\'': ending = '\''; break;
  case '\"': ending = '\"'; break;
  default:fdie(f,"internal error in skip_string");
  }
  s++;  
  while (*s && (*s != ending)) s++;
  if (*s == ending) return s+1;
  fdie(f, "string constant exceeds line in '%s'", expr);
}

double calculate(double x1, char op, double x2, XFILE *f) {
  switch (op) {
  case PLUS: return x1 + x2;
  case MINUS: return x1 - x2;
  case MULTIPLY: return x1 * x2;
  case POWER: return pow(x1,x2);
  case DIVIDE: {
    if (fabs(x2) < 0.000000000000001) ferr(f, "an attempt to divide by zero: %f/%f", x1, x2)
    else return x1 / x2;
  }
  default:ferr(f, "unknown operand '%c'", op);
  }
  return 0;
}

void reduce(double *x, char *op, XFILE *f) {
    switch (op[1]) {
    case PLUS:;
    case MINUS:
        x[0] = calculate(x[0], op[0], x[1], f);
        op[0] = op[1];
        x[1] = x[2];
        break;
    case MULTIPLY:;
    case DIVIDE:
        x[1] = calculate(x[1], op[1], x[2], f);
        break;
    }
}

char *gettok(char *s, char **tok, char *str, int separators) {
  char *x, *tmp;
  skip(s,WHITESPACE);
  x = s;
  if (separators) {
    while (*x && !strchr(str, *x)) x++;
  } else {
    while (*x && strchr(str, *x)) x++;
  }
  tmp = x - 1;
  while (strchr(WHITESPACE, *tmp)) tmp--;
  tmp++;
  *tok = duptok(s, tmp);
  skip(x,WHITESPACE);
  debug_print(150, "gettok %s", *tok);
  return x;
}

char *skip_symbol(char *s, XFILE *f) {
  char *tmp;
  while (*s && strchr(SYMBOL_CHARS, *s)) s++;
  if (*s == '[') {
    try(tmp = scan_closing(s, f));
    s = tmp + 1;
  }
  return s;
}

/* s points to a number, dot or minus
 * 
 */
char *skip_number(char *s) {
  if (*s == MINUS) s++;
  while (*s && strchr(NUMBERS, *s)) s++;
  if (*s == DOT) { /* skip decimals */
    s++;
    if (*s == DOT) { /* two dots */
      s--;
      return s;
    }
    while (*s && strchr(NUMBERS, *s)) s++;
  }
  return s;
}

char *get_number(char *s, char **number) {
  char *tmp = s, c;
  s = skip_number(s);
  c = *s;
  *s = '\0';
  try(*number = STRDUP(tmp));
  *s = c;
  return s;
}

/* expr = (symbol | number | '('expr')') {operand (symbol | number  | '('expr')')} 
*/
char *skip_expr(char *s, XFILE *f) {
  char *tmp, *prev_token = s;
  while (TRUE) {
    skip(s,WHITESPACE);
    switch (tt(*s)) {
    case PARENLEFT: /* it's a recursion */
      try(tmp = scan_closing(s, f));
      s = tmp + 1;
      break;
    case MINUS:;
    case DOT:;
    case NUMBER: /* it's a number */
      s = skip_number(s);
      break;
    case LETTER: /* it's a symbol */ 
      try(s = skip_symbol(s, f));
      break;
    default:
      fdie(f, "expected a number, opening paren, or a symbol after '%s'", prev_token);
    }
    skip(s,WHITESPACE);
    if (!*s) return s;
    prev_token = s;
    if (strchr(OPERANDS, *s)) s++; else return s;    
  }
}

eval_symbol_fct_type *eval_symbol_fct = eval_symbol;

void set_eval_symbol_fct(eval_symbol_fct_type *fct) {
  if (!fct) 
    eval_symbol_fct = eval_symbol;
  else 
    eval_symbol_fct = fct;
}

char *eval_expr(char *s, double *value, CMAP *C, INDEX *I, XFILE *f) {
  int i = 0;
  double x[3];
  char op[2] = {'\0', '\0'}, *tmp, tmpc, *prev_token = s;
  *value = 0;
  if (!s) die("?evaluate a null expression?");
  debug_print(165, "eval_expr %s", s);
  while (TRUE) {
    skip(s,WHITESPACE);    
    switch (tt(*s)) {
    case PARENLEFT: /* it's a recursion */
      try(tmp = scan_closing(s, f));      
      tmpc = *tmp;
      *tmp = '\0';
      s++;
      try(eval_expr(s, &(x[i]), C, I, f));
      *tmp = tmpc;
      s = tmp + 1;
      break;
    case MINUS:;
    case DOT:;
    case NUMBER: /* it's a number */
      tmp = skip_number(s);
      if (tmp == s) die("expected a number here '%s'", s);
      tmpc = *tmp;
      *tmp = '\0';
      try(scand(s, &x[i], f));
      *tmp = tmpc;
      s = tmp;
      break;
    case LETTER: /* it's a symbol */ 
      tmp = skip_symbol(s, f);
      tmpc = *tmp;
      *tmp = '\0';
      try(s = eval_symbol_fct(s, &(x[i]), C, I, f));
      *tmp = tmpc;
      s = tmp;
      break;
    case '\'':
    case '\"': /* it's a string */ 
      try(tmp = skip_string(s, f));      
      tmpc = *tmp;
      *tmp = '\0';      
      try(s = eval_symbol_fct(s, &(x[i]), C, I, f));
      *tmp = tmpc;
      s = tmp;
      break;
    default:
      fdie(f,"expected a number, opening paren, or a symbol at '%s'", s);
    }
    prev_token = s;
    skip(s,WHITESPACE);
    if (!*s) break;
    if (strchr(OPERANDS, *s)) {
      if (i == 2) {
	reduce(x, op, f);
	i = 1;
      }
      if (*s == MULTIPLY) {
	s++;
	if (*s == MULTIPLY) {
	  op[i] = POWER;
	  s++;
	} else 	
	  op[i] = MULTIPLY;
      } else {
	op[i] = *s;
	s++;
      }
      i++;                
    } else break;    
  }
  switch (i) {
  case 0:
    *value = x[0];
    break;
  case 1:
    *value = calculate(x[0], op[0], x[1], f);
    break;
  case 2:
    reduce(x, op, f);
    *value = calculate(x[0], op[0], x[1], f);
  }
  skip(s,WHITESPACE);
  return s;
}

char *str_add_index(char *name, int index) {
  char itoas[20];
  sprintf(itoas, "_%i", index);
  try(name = REALLOC(name, strlen(name)+strlen(itoas)+1));
  return strcat(name, itoas);
}

char *strip_indexes(char *s, int **indexes, int *n, XFILE *f) {
  char *x = s, tmpc, *buf;
  int i, d;
  *indexes = NULL;
  *n = 0;
  skip(s,WHITESPACE);
  while (*x && !strchr("_"WHITESPACE, *x)) x++;
  d = (*x == '_');
  tmpc = *x;
  *x = '\0';
  try(buf = STRDUP(s));
  *x = tmpc;
  while (d) {
    x++;
    s = x;
    while (*x && !strchr("_"WHITESPACE, *x)) x++;
    d = (*x == '_');
    tmpc = *x;
    *x = '\0';
    try(scani(s, &i, f));
    ++*n;
    try(*indexes = (int *)REALLOC(*indexes, *n * sizeof(int)));
    (*indexes)[*n - 1] = i;
    *x = tmpc;
  }
  return buf;
}

/* evaluates indexes and returns an expanded name */
char *parse_symbol(char *s, char **symbol, CMAP *C, INDEX *I, XFILE *f) {
  char *expr = s;
  double i_value;
  debug_print(170, "parse symbol %s", s);
  try(s = gettok(s, symbol, SYMBOL_CHARS, FALSE));
  debug_print(180, "parse symbol %s", s);
  if (*s == BRACKETLEFT) {
    s++;
    do {
      skip(s,WHITESPACE);      
      try(s = eval_expr(s, &i_value, C, I, f));
      *symbol = str_add_index(*symbol, round(i_value));
      skip(s,WHITESPACE);      
      if (*s == BRACKETRIGHT) {
	s++;
	return s;
      }
      if (*s == ',') 
	s++;
      else
	fdie(f,"in %s: expected a ',' or ']'", expr);
    } while (TRUE);
  }
  return s;
}

char *eval_symbol(char *s, double *value, CMAP *C, INDEX *I, XFILE *f) {
  char *name;
  CONSTANT *c = NULL;
  INDEX *i = NULL;
  try(s = parse_symbol(s, &name, C, I, f));
  debug_print(170, "eval symbol %s", name);
  if (C) c = search_constant(C, name);
  if (c  && c->has_value) {
    *value = c->value;
    return s;
  }
  if (I) i = search_index(I, name);
  if (i && i->has_value) {
    *value = i->value;
    return s;
  }
  fdie(f,"symbol '%s' has no value", name);
}

/* (expr (':'|"..") expr) | expr */
char *parse_range(char *s, char **min, char **max, XFILE *f) {
  char *tmp, tmpc;
  tmp = skip_expr(s, f);
  tmpc = *tmp;
  *tmp = '\0';
  try(*min = STRDUP(s));
  *tmp = tmpc;
  s = tmp;
  if ((*s == COLON) || ((*s == DOT) && (*(s++) == DOT))) {
    s++;
    tmp = s;
    tmp = skip_expr(tmp, f);
    tmpc = *tmp;
    *tmp = '\0';
    try(*max = STRDUP(s));
    *tmp = tmpc;
    s = tmp;
  } else
    try(*max = STRDUP(*min));
  return s;
}

char *eval_range(char *s, int *min, int *max, CMAP *C, INDEX *I, XFILE *f) {
  char *smin, *smax;
  double value;
  debug_print(170, "eval_range %s", s);
  try(s = parse_range(s, &smin, &smax, f));
  try(eval_expr(smin, &value, C, I, f));
  *min = round(value);
  try(eval_expr(smax, &value, C, I, f));
  *max = round(value);
  free(smin);
  free(smax);
  return s;
}

/* name ['=' range] */
char *parse_index(char *s, INDEX **i, CMAP *C, INDEX *I, XFILE *f) {
  debug_print(160, "parsing index decl. %s", s);
  try(*i = new_index(NULL));
  try(s = gettok(s, &((*i)->name), SYMBOL_CHARS, FALSE));
  if (!*s || (*s == COMMENT)) return s;
  if (*s != '=') fdie(f, "expected '=' after %s", (*i)->name);
  s++;
  try(s = eval_range(s, &((*i)->min), &((*i)->max), C, I, f));
  (*i)->has_bounds = TRUE;
  return s;
}

/* range | name ['=' range] {, range | name ['=' range]} */
char *parse_indexes(char *s, INDEX **NewI, int *n, CMAP *C, INDEX *I, XFILE *f) {
  INDEX *i, *tmpi;
  char *expr = s, *tmp, *s2;
  *n = 0;
  do {    
    skip(s,WHITESPACE);
    if (!*s) fdie(f,"expected index near '%s'", expr);
    try(i = new_index(NULL));
    tmp = s;
    try(s = gettok(s, &(i->name), ",=", TRUE));
    debug_print(100, "index in list %s", i->name);
    if ((tmpi = search_index(I, i->name))) {
      if (tmpi->has_bounds) {
	i->min = tmpi->min;
	i->max = tmpi->max;
	i->has_bounds = TRUE;
      }
      if (*s == EQUAL) {
	s++;
	try(s = parse_range(s, &(i->smin), &(i->smax), f));
	i->has_bounds = FALSE;
      }      
    } else {
      s2 = s;
      s = tmp;
      try(s = parse_range(s, &(i->smin), &(i->smax), f));
      i->has_bounds = FALSE;
    }
    *NewI = add_index(*NewI, i);
    ++*n;
    skip(s,WHITESPACE);
    if (!*s) return s;
    if (*s != ',') fdie(f,"expected ',' after '%s'", expr);
    s++;       
  } while (TRUE);
}

int init_indexes(int n, CMAP *C, INDEX *I, XFILE *f) {
  INDEX *root = I;
  int i, j;
  double value;
  char *s;
  for (i = n - 1; i >= 0; i--) {
    I = root;
    j = 0;
    while (I && (j < i)) {      
      I = I->next;
      j++;
    }
    if (!I->has_bounds) {
      s = I->smin;
      try(s = eval_expr(s, &value, C, I->next, f));
      I->min = value;
      s = I->smax;
      try(s = eval_expr(s, &value, C, I->next, f));
      I->max = value;
    }
    I->has_value = TRUE;
    I->value = I->min;
  }
  return TRUE;
}

int is_loop(int n, INDEX *I) {
  int i = 0;
  while (I && (i < n)) {
    if (I->min > I->max) return FALSE;
    I = I->next;
    i++;
  }
  return TRUE;
}

int loop(int n, CMAP *C, INDEX *I, XFILE *f) {
  INDEX *root = I;
  int i = 0;
  while (I && (i < n)) {
    if (I->min <= I->max) I->value++;
    if ((I->min > I->max) || (I->value > I->max)) {
      I = I->next;
      i++;
    } else {      
      return init_indexes(i, C, root, f);
    }
  }
  return FALSE;
}

int test_condition(char *condition, int *result, CMAP *C, INDEX *I, XFILE *in) {
  double x, y;  
  char test_c, *tmp;
  try(condition = eval_expr(condition, &x, C, I, in));
  if(*condition) {
    test_c = *condition;
    condition++;
    try(tmp = eval_expr(condition, &y, C, I, in));
    if (*tmp) fdie(in, "extra input in condition '%s'", tmp);
    switch (test_c) {
    case '<':*result = x < y; break;
    case '>':*result = x > y; break;
    case '=':*result = x == y; break;
    default:fdie(in, "parse error in '%s'", condition);
    }
  } else *result = round(x);
  return TRUE;
}

/* name
 * name = expr
 * name'['indexes']' 
 * name'['indexes']' = '{' value {,value} '}'
 * If (<condition>; <constant_declaration>)
 */
char *parse_constant(char *s, CMAP *NewC, CMAP *C, INDEX *I, XFILE *in) {
  CONSTANT *c;
  INDEX *LocalI = NULL, *ix;
  int i, j, n, table_initialization, fct_nr, test = TRUE;
  char *expr = s, *tmp, tmpc, *tmp2, *end_b, *end_c, *cond = NULL;
  char *token;
  debug_print(4, "parsing constant decl. %s", s);
  if (!*s) return s;
  try(s = gettok(s, &token, SYMBOL_CHARS, FALSE));
  n = N_FCT_TAGS;
  fct_nr = getcnr(token, fct_tags, fct_tag_numbers, &n);
  if (fct_nr) {
    if (fct_nr == IF_FCT) {
      if (*s != '(') cfdie(in, "expected '(' in a function");
      try(end_b = scan_closing(s, in));
      *end_b = '\0';
      s++;
      skip(s,WHITESPACE);
      cond = s;
      while (*s && (*s != ';')) s++;
      if (*s != ';') cfdie(in, "expected ';' in a function");
      *s = '\0';
      end_c = s;
      s++;
      free(token);
      try(s = gettok(s, &token, SYMBOL_CHARS, FALSE));
    } else {
      cfdie(in, "function %s not supported here", token);
    }
  }
  if (*s != '[') {
    try(c = new_constant(token));
    if (*s == '=') {
      s++;
      try(s = eval_expr(s, &(c->value), C, NULL, in));
      c->has_value = TRUE;
    }
    add_constant(NewC, c);
    if (fct_nr == IF_FCT) {
      *end_b = ')';
      *end_c = ';';
    }
    return s;
  }
  ctry(tmp = scan_closing(s, in));
  tmpc = *tmp;
  *tmp = '\0';
  s++;
  ctry(s = parse_indexes(s, &LocalI, &n, C, I, in));
  *tmp = tmpc;
  s = tmp + 1;
  skip(s,WHITESPACE);
  table_initialization = (*s == '=');
  if (table_initialization) {
    s++;
    skip(s,WHITESPACE);
    if (*s != '{') fdie(in,"expected '{' near '%s'", expr);
    ctry(tmp = scan_closing(s, in));
    tmpc = *tmp;
    *tmp = '\0';
    ++s;
    while (*s && strchr(WHITESPACE, *s)) ++s;
  }
  if (init_indexes(n, C, LocalI, in)) {
    do {
      if (is_loop(n, LocalI)) {
	if (cond) ctry(test_condition(cond, &test, C, LocalI, in));
	if (test) {
	  ctry(c = new_constant(token));
	  for (i = 0; i < n; i++) {
	    ix = LocalI;
	    for (j = i + 1; j < n; j++) ix = ix->next;
	    c->name = str_add_index(c->name, ix->value);
	  }
	  if (table_initialization) {
	    tmp2 = s;
	    ctry(s = eval_expr(s, &(c->value), C, LocalI, in));
	    c->has_value = TRUE;
	    skip(s,WHITESPACE);
	    if (!*s) s = tmp2; 
	    else {
	      if (*s != ',') fdie(in,"expected ',' after %s", expr);
	      s++; 
	    }	  
	  }
	  add_constant(NewC, c);
	}
      }
    } while (loop(n, C, LocalI, in));
  }
  if (table_initialization) {
    *tmp = tmpc;
    s = tmp + 1;
  }
  if (fct_nr == IF_FCT) {
    *end_b = ')';
    *end_c = ';';
  }
  return s;
catch:
  free(token);
  return NULL;
}


/*---------------------------------------------------------------------------*/


/* name ['=' range] {, name ['=' range]} */
char *parse_function_indexes(char *s, INDEX **NewI, int *n, CMAP *C, INDEX *I, XFILE *f) {
  INDEX *i, *tmpi;
  char *tmp;
  *n = 0;
  skip(s,WHITESPACE);
  if (*s == '[') s++; /* old format */
  do {
    skip(s,WHITESPACE);
    try(i = new_index(NULL));
    tmp = s;
    try(s = gettok(s, &(i->name), SYMBOL_CHARS, FALSE));    
    debug_print(180, "new function counter %s", i->name);
    if (*s == '=') {
      s++;
      try(s = parse_range(s, &(i->smin), &(i->smax), f));
      i->has_bounds = FALSE;      
    } else {
      if (!(tmpi = search_index(I, i->name))) fdie(f,"undeclared index '%s'", i->name);
      i->min = tmpi->min;
      i->max = tmpi->max;
      i->has_bounds = TRUE;
    }
    *NewI = add_index(*NewI, i);
    ++*n;    
    if (*s == ']') {s++; skip(s,WHITESPACE);} /* old */
    if (!*s) return s;
    if (*s != ',') fdie(f,"extra input after %s", i->name);
    s++;       
  } while (TRUE);
}

char *parse_function(char *s, int fct_nr, CMAP *C, CMAP *variables, INDEX *I, INDEX *local_indexes, int *did_gen_out, XFILE *in, XFILE *out) {
  char *tmp1, tmpc1, *tmp2, tmpc2, *local, char_between = '\0';
  int n, dgo, test;
  try(tmp1 = scan_closing(s, in));
  tmpc1 = *tmp1;
  *tmp1 = '\0';
  s++;
  skip(s,WHITESPACE);
  local = s;
  while (*s && (*s != ';')) s++;
  if (*s != ';') fdie(in, "expected ';' in a function");
  tmp2 = s;
  tmpc2 = *tmp2;
  *tmp2 = '\0';
  s++;
  switch (fct_nr) {
  case IF_FCT:
    try(test_condition(local, &test, C, local_indexes, in));
    if (test) try(parse_block(s, C, variables, I, local_indexes, did_gen_out, in, out));
    break;
  case SUM_FCT:
    char_between = '+'; 
  case FORALL_FCT:
    if (!char_between) char_between = ';';
    try(local = parse_function_indexes(local, &local_indexes, &n, C, I, in));
    if (*local) fdie(in, "parse error in '%s'", local);
    if (init_indexes(n, C, local_indexes, in)) {
      do {
	if (is_loop(n, local_indexes)) {	  
	  try(parse_block(s, C, variables, I, local_indexes, &dgo, in, out));
	  if (dgo) {
	    stream_out_char(out, char_between);
	    *did_gen_out = TRUE;
	  }
	}
      } while (loop(n, C, local_indexes, in));
    }
    remove_index(&local_indexes, n);
  }
  *tmp2 = tmpc2;
  *tmp1 = tmpc1;
  return tmp1 + 1;
}

/* { (expr) | operand | '=','<','>',';' 
   | int | variable | fct'('local; block')' } */
char *parse_block(char *s, CMAP *C, CMAP *variables, INDEX *I, INDEX *local_indexes, int *did_gen_out, XFILE *in, XFILE *out) {
  char *token, *tmp, tmpc;
  char *name;
  CONSTANT *c;
  INDEX *i;
  double value;
  int n;
  CNR fct_nr;
  *did_gen_out = FALSE;
  while(TRUE) {
    skip(s,WHITESPACE);
    if (!*s || (*s == '#')) return s;
    if (*s == '(') {
      try(tmp = scan_closing(s, in));      
      tmpc = *tmp;
      *tmp = '\0';
      s++;
      try(s = eval_expr(s, &value, C, local_indexes, in));
      *tmp = tmpc;
      s = tmp + 1;
      stream_out_number(out, value);
      *did_gen_out = TRUE;
    } else if (strchr(OPERANDS, *s)) {
      stream_out_char(out, *s);
      s++;
      *did_gen_out = TRUE;
    } else if (strchr("=<>;", *s)) {
      stream_out_char(out, *s);
      s++;
      *did_gen_out = TRUE;
    } else if (strchr(".0123456789", *s)) {
      try(s = get_number(s, &token));
      try(scand(token, &value, NULL));
      stream_out_number(out, value);
      *did_gen_out = TRUE;
    } else {
      tmp = gettok(s, &token, SYMBOL_CHARS, FALSE);
      debug_print(180, "block starts with token %s", token);
      n = N_FCT_TAGS;
      fct_nr = getcnr(token, fct_tags, fct_tag_numbers, &n);
      if (fct_nr) s = tmp;
      if (fct_nr == INT_OP) {
        stream_out_token(out, INT_TAG);
        stream_out_char(out, ' ');
        *did_gen_out = TRUE;
      } else if (!fct_nr) {
	try(s = parse_symbol(s, &name, C, local_indexes, in));
	if (C) c = search_constant(C, name); else c = NULL;
	if (c && c->has_value) stream_out_number(out, c->value);
	else {
	  i = search_index(local_indexes, name);
	  if (i && i->has_value) stream_out_number(out, i->value);
	  else {
	    if (C) c = search_constant(variables, name); else c = NULL;
	    if (!c && verbose) ferr(in,"undeclared variable '%s'",name); 
	    stream_out_token(out, name);
	    if (strcmp(name, "int") == 0) stream_out_char(out, ' ');
	  }
	}
	*did_gen_out = TRUE;
      } else {
	if (*s != '(') fdie(in, "expected '(' after '%s'", token);
	try(s = parse_function(s, fct_nr, C, variables, I, local_indexes, did_gen_out, in, out));
      }
    }
  }
}





