/*

    this source file is a part of 
    lp_parse - a preprocessor for MIP models
    Copyright (C) 1996  Ari Jolma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    for more information please contact 
    Ari Jolma
    Helsinki University of Technology
    Laboratory of Water Resources Engineering
    Building W
    02150 Espoo, Finland
    E-mail: Ari.Jolma@hut.fi

 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include "util.h"

char *pn = NULL;
int verbose = FALSE;

char *strend(char *s) {
  while (*s) s++;
  return s;
}

char *trim(char *s) {
  char *e = strend(s);
  if (e != s) {
    e--;
    while (strchr(WHITESPACE, *e)) e--;
    e++;
    *e = '\0';
  }
  return s;
}

/* --------------------------------------------------------------------------*/


void *my_malloc(size_t size) {
  void *ptr;  
  ptr = malloc(size);
  debug_print(500, "malloc %X", (int)ptr);
  if (!ptr && (size > 0)) die("Out of memory (%i)", size);
  return ptr;
}

void my_free(void *ptr) {
  debug_print(500, "free %X", (int)ptr);
  free(ptr);
}

void *my_realloc(void *ptr, size_t size) {
  void *tmp;  
  tmp = realloc(ptr, size);
  debug_print(500, "realloc %X %X", (int)ptr, (int)tmp);
  if (!tmp && (size > 0)) die("Out of memory (%i)", size);
  return tmp;
}

void *my_calloc(size_t nmemb, size_t size) {
  void *tmp;
  tmp = calloc(nmemb, size);
  debug_print(500, "calloc %X", (int)tmp);
  if (!tmp && (nmemb*size > 0)) die("Out of memory (%li)", (long)nmemb*size);
  return tmp;
}

char *my_strdup(const char *c) {
  char *t;
  t = strdup(c);
  if (!t) die("Out of memory");
  debug_print(500, "strdup %X", (int)t);  
  return t;
}

char *duptok(char *beg, char *end) {
  char *dup, c = *end;
  *end = '\0';
  try(dup = STRDUP(beg));
  *end = c;
  return dup;
}

int scani(const char *str, int *i, XFILE *f) {
  if (str && (sscanf(str, "%i", i) < 1)) fdie(f,"'%s' is not an integer",str);
  return TRUE;
}

int scand(const char *str, double *d, XFILE *f) {
  debug_print(250, "scand %s", str);
  if (str && (sscanf(str, "%lf", d) < 1)) fdie(f,"'%s' is not a number",str);
  return TRUE;
}


/* --------------------------------------------------------------------------*/


/* class_number is a string of numbers, the rightmost is the most significant
   this function tests if sub is a descendant of super */

#define CLASS_NUMBER_BASE 0x0010

int is_sub_class(CNR super, CNR sub) {
  int i = 1;
  while (super) {
    if ((super % CLASS_NUMBER_BASE) != (sub % CLASS_NUMBER_BASE)) return FALSE;
    super /= CLASS_NUMBER_BASE;
    sub /= CLASS_NUMBER_BASE;
  }
  while (sub) {
    sub /= CLASS_NUMBER_BASE;
    i++;
  }
  return i;
}

CNR getcnr(char *cn, char **cns, CNR *cnrs, int *n) {
  do {
    (*n)--;
    if (!strcmp(cn, cns[*n])) 
      return cnrs[*n];
  } while (*n > 0);
  return 0;
}

char *getcn(CNR cnr, char **cns, CNR *cnrs, int *n) {
  do {
    (*n)--;
    if (cnr == cnrs[*n])
      return cns[*n];
  } while (*n > 0);
  return NULL;
}


/* --------------------------------------------------------------------------*/


char xeof[2] = {XEOF,'\0'};

char *xgetline(XFILE *f) {
  register int i = 0, j;
  char *buf = NULL, *ret = xeof;
  int buf_size = 256;
  if (f->f == NULL) {    
    if (f->l < f->lines) {
      i = f->l;
      f->ln = f->line_nrs[i];
      ret = f->data[i];
    }
  } else if (feof(f->f)) {
  } else {    
    try(buf = (char *)MALLOC(buf_size));
    while (TRUE) {
      if (feof(f->f)) {
	buf[i] = '\0'; 
	break;
      }
      buf[i] = (char)fgetc(f->f);
      if ((buf[i] == 4) || (buf[i] == -1)) {
	buf[i] = '\0'; 
	break;
      }    
      else if (buf[i] == '\n') {
	f->ln++;
	j = i;
	while ((j >= 0) && strchr(WHITESPACE, buf[j])) j--;
	if ((buf[j] == '\\') || (buf[j] == '_')) 
	  i = j;
	else {
	  buf[i] = '\0'; 
	  break;
	}
      } else {
	i++;
	if (i == buf_size) {
	  buf_size += 256;
	  try(buf = (char *)realloc(buf, buf_size));
	}
      }
    }
    if (f->in_include>0) f->line_nrs[f->l] = f->in_include; 
    else {
      if (f->line_nrs) f->line_nrs[f->l] = f->ln;
    }    
    ret = buf;
  }
  f->l++;
  if ((strncmp(ret,"#eof",4) == 0) || (strncmp(ret,"__END__",7) == 0)) {
    if (buf) free(buf);
    return xeof;
  }
  return ret;
}


/* --------------------------------------------------------------------------*/


char *pop_filename(char *path) {
  char *tmp = path;
  if (path == NULL) return NULL;
  while (*path) path++;
  while ((path != tmp) && (*path != '/')) path--;
  if (*path == '/') path++;
  return path;
}

XFILE *new_XFILE(char *name) {
  XFILE *xfile;
  try(xfile  = (XFILE *)MALLOC(sizeof(XFILE)));
  xfile->name = NULL; 
  xfile->f = NULL;
  xfile->ln = 0;
  xfile->l = 0;
  xfile->lines = 0;
  xfile->line_nrs = NULL;
  xfile->data = NULL;
  xfile->buffer = NULL;
  xfile->buf_size = 0;
  xfile->in_include = 0;
  if (name) ctry(xfile->name = STRDUP(name));
  return xfile;
catch:
  free(xfile);
  return NULL;
}

XFILE *xstdin() {
  XFILE *xfile = new_XFILE(NULL);  
  if (!xfile) return NULL;
  xfile->f = stdin;
  return xfile;
}

XFILE *xstdout() {
  XFILE *xfile = new_XFILE(NULL);  
  if (!xfile) return NULL;
  xfile->f = stdout;
  return xfile;
}

XFILE *xbuffer(size_t size) {
  XFILE *xfile = new_XFILE(NULL);  
  if (!xfile) return NULL;
  xfile->buffer = MALLOC(size);
  if (!xfile->buffer) return NULL;
  return xfile;
}

int input_file(XFILE *xfile, char *name);

int input_line(XFILE *xfile) {
  char *line,*path,*tmp;
  int tmpl,itmp;
  while (TRUE) {
    try(line = xgetline(xfile));
    if (strncmp(line,"#include",8) == 0) {
      xfile->l--;
      line += 8;
      line = strtok(line, WHITESPACE);
      try(path = MALLOC(strlen(xfile->name)+strlen(line)+1));
      strcpy(path,xfile->name);
      tmp = pop_filename(path);
      strcpy(tmp,line);
      tmpl = xfile->ln;
      xfile->ln = 0;
      itmp = xfile->in_include;
      xfile->in_include = tmpl;
      try(input_file(xfile, path));
      xfile->in_include = itmp;
      xfile->ln = tmpl;
      if (path) free(path);
    } else break;
  }
  xfile->data[xfile->l-1] = line;
  debug_print(50, "%i %s", xfile->l, line);
  if (xfile->l == xfile->lines) {
    xfile->lines += 256;
    try(xfile->data  = (char **)REALLOC(xfile->data, xfile->lines*sizeof(char *)));
    try(xfile->line_nrs  = (int *)REALLOC(xfile->line_nrs, xfile->lines*sizeof(int)));
  }
  return TRUE;
}

int input_file(XFILE *xfile, char *name) {
  int ret = FALSE, tmp;
  FILE *saved = xfile->f;
  xfile->f = fopen(name, "r");
  if (!(xfile->f)) {
    if (!xfile->ln && !xfile->in_include) 
      err("'%s': file not found", name);
    else {
      tmp = xfile->ln;
      xfile->ln = xfile->in_include;
      ferr(xfile, "'%s': file not found", name);
      xfile->ln = tmp;
    }
    goto exit2;
  } 
  do {
    if (!input_line(xfile)) goto exit1;
  } while (!feof(xfile->f));
  ret = TRUE;
exit1:
  fclose(xfile->f);
exit2:
  xfile->f=saved;
  return ret;
}

XFILE *open_file(char *name, char *flags) {  
  XFILE *xfile;  
  try(xfile = new_XFILE(name));
  if (flags[0] != 'r') {
    xfile->f = fopen(name, flags);
    if (!(xfile->f)) die("can't open file '%s' for writing", name);
    return xfile;
  } else {
    xfile->lines = 256;
    try(xfile->data  = (char **)MALLOC(xfile->lines * sizeof(char *)));
    try(xfile->line_nrs  = (int *)MALLOC(xfile->lines * sizeof(int)));
    try(input_file(xfile, name));
    xfile->lines = xfile->l;
    xfile->ln = 0;
    xfile->l = 0;
    return xfile;
  }
}

char *xfn(XFILE *xf) {
  try(xf);
  return xf->name;
}

void xclose(XFILE *x) {
  if (x) {
    if ((x->f != stdin) && (x->f != stdout)) fclose(x->f);
    if (x->name) free(x->name);
    if (x->line_nrs) free(x->line_nrs);
    if (x->data) free(x->data);
    free(x);
  }
}

