/*

    this source file is a part of 
    lp_parse - a preprocessor for MILP models
    Copyright (C) 1996  Ari Jolma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

    for more information please contact 
    Ari Jolma
    Helsinki University of Technology
    Laboratory of Water Resources Engineering
    Building W
    02150 Espoo, Finland
    E-mail: Ari.Jolma@hut.fi

 */

#define TRUE 1
#define FALSE 0

#define MALLOC my_malloc
#define REALLOC my_realloc
#define CALLOC my_calloc
#define STRDUP my_strdup
#define FREE my_free

char *duptok(char *beg, char *end);

void *my_malloc(size_t size);
void my_free(void *ptr);
void *my_realloc(void *ptr, size_t size);
void *my_calloc(size_t nmemb, size_t size);
char *my_strdup(const char *c);

#define LETTERS "abcdefghijklmnopqrstuvwxyzABCDEFGHIJKLMNOPQRSTUVWXYZ"
#define NUMBERS "0123456789"
#define MINUS '-'
#define PLUS '+'
#define MULTIPLY '*'
/* ** is substituted internally */
#define POWER '^'
#define DIVIDE '/'
#define OPERANDS "+-*^/"
#define DOT '.'
#define UNDERSCORE '_'
#define PARENLEFT '('
#define BRACKETLEFT '['
#define PARENRIGHT ')'
#define BRACKETRIGHT ']'
#define COLON ':'
#define COMMENT '#'
#define EQUAL '='

#define LETTER -1
#define NUMBER -2
#define SYMBOL_CHARS LETTERS"_"NUMBERS
#define PARENS "{([])}"
#define RIGHTPARENS "])}"

#define WHITESPACE " \t\r\n"

extern char *pn; /* program's name */
extern int verbose;

#ifdef DEBUG
#define debug_print(level,msg,x...) {if (verbose>level) \
  printf("%s: "msg"\n", pn , ## x);}
#else
#define debug_print(level,msg,x...) {;}
#endif

#define err(msg,x...) fprintf(stderr, "%s: "msg"\n", pn , ## x)

#define die(msg,x...) {err(msg, ## x); return 0;}
#define cdie(msg,x...) {err(msg, ## x); goto catch;}

#define try(x) {if (!(x)) return 0;}
#define ctry(x) {if (!(x)) goto catch;}
#define mtry(x) {if (!(x)) return 1;}

#define ferr(f,msg,x...) {if ((f)) \
  fprintf(stderr,"%s:%s:%i: "msg"\n", pn, pop_filename(f->name), \
	  f->ln , ## x); else fprintf(stderr, "%s: "msg"\n", pn , ## x);}

#define fdie(f,msg,x...) {ferr(f,msg, ## x) return 0;}
#define cfdie(f,msg,x...) {ferr(f,msg, ## x) goto catch;}

#define min(a,b) ((a) < (b) ? (a) : (b))
#define max(a,b) ((a) > (b) ? (a) : (b))

#define round(x) ( (long) ((x)+0.5) )

#define sqr(a) ((a) * (a))

#define interpolate(x1,x,x2,y1,y2) ((y1)+((x)-(x1))*((y2)-(y1))/((x2)-(x1)))

#define swap(a,b,tmp) {(tmp) = (a); (a) = (b); (b) = (tmp);}

#define sort(table, n, cmp) do {f = FALSE; for(i=0;i<(n)-1;i++) \
  {if ((cmp)) {swap((table)[i],(table)[i+1],tmp);f = TRUE;}}} while(f);

#define skip(s,chars) while (*(s) && strchr((chars), *(s))) (s)++

char *strend(char *s);
char *trim(char *s);


/* --------------------------------------------------------------------------*/

#define CNR unsigned long

CNR getcnr(char *cn, char **cns, CNR *cnrs, int *n);
char *getcn(CNR cnr, char **cns, CNR *cnrs, int *n);

int is_sub_class(CNR super, CNR sub);

/* --------------------------------------------------------------------------*/


struct xfile {
  FILE *f;
  char *name;
  int ln;      /* current line */
  int l;
  int lines;   /* number of lines */
  int *line_nrs;
  char **data;
  char *buffer;
  size_t buf_size;
  int in_include;
};

#define XFILE struct xfile

XFILE *new_XFILE(char *name);

#define XEOF '\4'

/* f is either a stream (reads a line from it) or 
   a string table (returns a pointer to the next unread line)
   returns NULL if error, 
   returns a pointer to XEOF if eof (no more lines) */

char *xgetline(XFILE *f);


/* --------------------------------------------------------------------------*/


char *pop_filename(char *path);

XFILE *xstdin();
XFILE *xstdout();
XFILE *xbuffer(size_t size);

XFILE *open_file(char *name, char *flags);
void xclose(XFILE *x);


/* 
 * x is the input file, 
 * whose name and current input line number is used
 * in an error message in the case of an error 
*/

int scani(const char *str, int *i, XFILE *x);
int scand(const char *str, double *d, XFILE *x);






